from multifruits import extract_filename


def test_extract_filename():
    assert extract_filename({b'filename': b'foo.bar'}) == 'foo.bar'


def test_extract_filename_without_filename():
    assert extract_filename({}) == ''


def test_extract_filename_star():
    assert extract_filename({b'filename*': b"utf-8''foo.bar"}) == 'foo.bar'


def test_extract_filename_and_star():
    assert extract_filename({
        b'filename*': b"utf-8''foo.bar",
        b'filename': b"baz.quux"
    }) == 'foo.bar'


def test_extract_filename_and_star_wrong_encoding():
    assert extract_filename({
        b'filename*': b"unknown''foo.bar",
        b'filename': b"baz.quux"
    }) == 'baz.quux'


def test_extract_filename_star_wrong_encoding():
    assert extract_filename({b'filename*': b"unknow''foo.bar"}) == 'foo.bar'


def test_extract_filename_star_latin1_with_accent():
    assert extract_filename({
        b'filename*': "iso-8859-1''foo-ä.html".encode('latin1')
    }) == 'foo-ä.html'


def test_extract_filename_star_utf8_with_accent():
    assert extract_filename({
        b'filename*': "UTF-8''foo-ä-€.html".encode()
    }) == 'foo-ä-€.html'


def test_extract_filename_star_no_encoding():
    assert extract_filename({
        b'filename*': "''foo-ä-€.html".encode()
    }) == "foo-ä-€.html"


def test_extract_filename_star_no_encoding_but_filename():
    assert extract_filename({
        b'filename*': "''foo-ä-€.html".encode(),
        b'filename': b"baz.quux"
    }) == "baz.quux"


def test_extract_filename_star_without_quotes():
    assert extract_filename({
        b'filename*': "foo-ä-€.html".encode()
    }) == "foo-ä-€.html"


def test_extract_filename_star_without_quotes_but_filename():
    assert extract_filename({
        b'filename*': "foo-ä-€.html".encode(),
        b'filename': b"baz.quux"
    }) == "baz.quux"



from collections import namedtuple

from multifruits import Parser
import pytest


class Handler:

    def __init__(self, boundary):
        self.parts = []
        self.parser = Parser(self, boundary)
        self.on_body_begin_called = 0
        self.on_body_complete_called = 0
        self.on_headers_complete_called = 0

    def feed_data(self, data):
        self.parser.feed_data(data)

    def on_body_begin(self):
        self.on_body_begin_called += 1
        print('on_body_begin')

    def on_part_begin(self):
        self._current = namedtuple('Data', ['headers', 'content'])
        self._current.headers = {}
        self._current.content = b''
        print('on_part_begin')

    def on_header(self, field, value):
        self._current.headers[field] = value
        print(field, value)

    def on_headers_complete(self):
        self.on_headers_complete_called += 1
        print('on_headers_complete')

    def on_data(self, data: bytes):
        self._current.content += data

    def on_part_complete(self):
        self.parts.append(self._current)
        self._current = None
        print('on_part_complete')

    def on_body_complete(self):
        self.on_body_complete_called += 1
        print('on_body_complete')


def test_parse():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=baz; filename="baz.png"\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'abcdef\r\n'
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name=baz; filename="baz.png"',
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[1].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 2


def test_parse_invalid_header_name_character():
    body = (b'--foo\r\n'
            b'Content->Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    with pytest.raises(ValueError):
        form.feed_data(body)


def test_parse_kind_of_boundary_in_data():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--fo\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts[0].content == b'abc\r\n--fo'


def test_parse_chunked_boundary():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body[:4])
    form.feed_data(body[4:])
    assert form.parts[0].content == b'abc'


def test_parse_chunked_last_boundary():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body[:-4])
    form.feed_data(body[-4:])
    assert form.parts[0].content == b'abc'


def test_parse_char_by_char():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    for idx in range(len(body)):
        form.feed_data(body[idx:idx+1])
    assert form.parts[0].content == b'abc'


def test_parse_missing_boundary():
    with pytest.raises(ValueError) as e:
        Handler(b'multipart/form-data')
    assert str(e.value) == 'Missing boundary in Content-Type.'


def test_parse_empty():
    body = b''
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts == []
    assert form.on_body_begin_called == 0
    assert form.on_body_complete_called == 0
    assert form.on_headers_complete_called == 0


def test_parse_empty_boundary():
    body = b'--bar\r\nbaz\r\n--bar--'
    form = Handler(b'multipart/form-data; boundary=')
    with pytest.raises(ValueError):
        form.feed_data(body)


def test_parse_preamble():
    body = (b'preamble'
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[0].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 1


def test_parse_epilogue():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--'
            b'epilogue')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[0].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 1


def test_parse_filename_star():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=baz; '
            b'filename="iso-8859-1\'\'baz-\xe9.png"\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'abcdef\r\n'
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts[0].headers == {
        b'Content-Disposition': (b'form-data; name=baz; '
                                 b'filename="iso-8859-1\'\'baz-\xe9.png"'),
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[1].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 2


def test_parse_feed_data_content_chunked():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=baz; filename="baz.png"\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'abc')
    body2 = (b'def\r\n'
             b'--foo\r\n'
             b'Content-Disposition: form-data; name="text1"\r\n'
             b'\r\n'
             b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.on_headers_complete_called == 1
    form.feed_data(body2)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name=baz; filename="baz.png"',
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[1].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 2


def test_parse_feed_data_header_name_chunked():
    body = (b'--foo\r\n'
            b'Content-Disposit')
    body2 = (b'ion: form-data; name=baz; filename="baz.png"\r\n'
             b'Content-Type: image/png\r\n'
             b'\r\n'
             b'abcdef\r\n'
             b'--foo\r\n'
             b'Content-Disposition: form-data; name="text1"\r\n'
             b'\r\n'
             b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.on_headers_complete_called == 0
    form.feed_data(body2)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name=baz; filename="baz.png"',
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[1].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 2


def test_parse_feed_data_header_value_chunked():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=baz; filename="ba')
    body2 = (b'z.png"\r\n'
             b'Content-Type: image/png\r\n'
             b'\r\n'
             b'abcdef\r\n'
             b'--foo\r\n'
             b'Content-Disposition: form-data; name="text1"\r\n'
             b'\r\n'
             b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.on_headers_complete_called == 0
    form.feed_data(body2)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name=baz; filename="baz.png"',
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[1].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 2


def test_parse_feed_data_boundary_chunked():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=baz; filename="baz.png"\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'abcdef\r\n'
            b'--f')
    body2 = (b'oo\r\n'
             b'Content-Disposition: form-data; name="text1"\r\n'
             b'\r\n'
             b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    form.feed_data(body2)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name=baz; filename="baz.png"',
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[1].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 2


def test_parse_two_parts_with_same_name():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=bar; filename=tmp.png\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'abcdef\r\n'
            b'--foo\r\n'
            b'Content-Disposition: form-data; name=bar; filename=foo.png\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'ghijkl\r\n'
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts[0].headers == {
        b'Content-Disposition': b'form-data; name=bar; filename=tmp.png',
        b'Content-Type': b'image/png'
    }
    assert form.parts[0].content == b'abcdef'
    assert form.parts[1].headers == {
        b'Content-Disposition': b'form-data; name=bar; filename=foo.png',
        b'Content-Type': b'image/png'
    }
    assert form.parts[1].content == b'ghijkl'
    assert form.parts[2].headers == {
        b'Content-Disposition': b'form-data; name="text1"'
    }
    assert form.parts[2].content == b'abc'
    assert form.on_body_begin_called == 1
    assert form.on_body_complete_called == 1
    assert form.on_headers_complete_called == 3


def test_parse_with_null_byte():
    body = (b'--foo\r\n'
            b'Content-Disposition: form-data; name=baz; filename="baz.png"\r\n'
            b'Content-Type: image/png\r\n'
            b'\r\n'
            b'abcdef\x00ghi\r\n'
            b'--foo\r\n'
            b'Content-Disposition: form-data; name="text1"\r\n'
            b'\r\n'
            b'abc\x00def\r\n--foo--')
    form = Handler(b'multipart/form-data; boundary=foo')
    form.feed_data(body)
    assert form.parts
    assert form.parts[0].content == b'abcdef\x00ghi'
    assert form.parts[1].content == b'abc\x00def'



# http://greenbytes.de/tech/tc2231/

from multifruits import parse_content_disposition


def test_parse_empty():
    dtype, params = parse_content_disposition(b'')
    assert dtype is None
    assert params == {}


def test_inlonly():
    dtype, params = parse_content_disposition(b'inline')
    assert dtype == b'inline'
    assert params == {}


def test_inlonlyquoted():
    dtype, params = parse_content_disposition(b'"inline"')
    assert dtype == b'inline'
    assert params == {}


def test_inlwithasciifilename():
    dtype, params = parse_content_disposition(b'inline; filename="foo.html"')
    assert dtype == b'inline'
    assert params == {b'filename': b'foo.html'}


def test_inlwithfnattach():
    dtype, params = parse_content_disposition(
        b'inline; filename="Not an attachment!"')
    assert dtype == b'inline'
    assert params == {b'filename': b'Not an attachment!'}


def test_attonly():
    dtype, params = parse_content_disposition(b'attachment')
    assert dtype == b'attachment'
    assert params == {}


def test_attonlyquoted():
    dtype, params = parse_content_disposition(b'"attachment"')
    assert dtype == b'attachment'
    assert params == {}


def test_attonlyucase():
    dtype, params = parse_content_disposition(b'ATTACHMENT')
    assert dtype == b'attachment'
    assert params == {}


def test_attwithasciifilename():
    dtype, params = parse_content_disposition(
        b'attachment; filename="foo.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.html'}


def test_inlwithasciifilenamepdf():
    dtype, params = parse_content_disposition(
        b'attachment; filename="foo.pdf"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.pdf'}


def test_attwithasciifilename25():
    dtype, params = parse_content_disposition(
        b'attachment; filename="0000000000111111111122222"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'0000000000111111111122222'}


def test_attwithasciifilename35():
    dtype, params = parse_content_disposition(
        b'attachment; filename="00000000001111111111222222222233333"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'00000000001111111111222222222233333'}


def test_attwithasciifnescapedchar():
    dtype, params = parse_content_disposition(
        br'attachment; filename="f\oo.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.html'}


def test_attwithasciifnescapedquote():
    dtype, params = parse_content_disposition(
        rb'attachment; filename="\"quoting\" tested.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'"quoting" tested.html'}


def test_attwithquotedsemicolon():
    dtype, params = parse_content_disposition(
        b'attachment; filename="Here\'s a semicolon;.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b"Here's a semicolon;.html"}


def test_attwithfilenameandextparam():
    dtype, params = parse_content_disposition(
        b'attachment; foo="bar"; filename="foo.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.html', b'foo': b'bar'}


# def test_attwithfilenameandextparamescaped():
#     dtype, params = parse_content_disposition(
#         b'attachment; foo="\"\\";filename="foo.html"')
#     assert dtype == b'attachment'
#     assert params == {b'filename': b'foo.html', b'foo': b'"\\'}


def test_attwithasciifilenameucase():
    dtype, params = parse_content_disposition(
        b'attachment; FILENAME="foo.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.html'}


def test_attwithasciifilenamenq():
    dtype, params = parse_content_disposition(
        b'attachment; filename=foo.html')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.html'}


def test_attemptyparam():
    dtype, params = parse_content_disposition(b'attachment; ;filename=foo')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo'}


def test_attwithasciifilenamenqws():
    dtype, params = parse_content_disposition(
        b'attachment; filename=foo bar.html')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo bar.html'}


def test_attwithfntokensq():
    dtype, params = parse_content_disposition(
        b"attachment; filename='foo.html'")
    assert dtype == b'attachment'
    assert params == {b'filename': b"'foo.html'"}


def test_attwithisofnplain():
    dtype, params = parse_content_disposition(
        'attachment; filename="foo-ä.html"'.encode())
    assert dtype == b'attachment'
    assert params == {b'filename': 'foo-ä.html'.encode()}


def test_attwithutf8fnplain():
    dtype, params = parse_content_disposition(
        'attachment; filename="foo-Ã¤.html"'.encode())
    assert dtype == b'attachment'
    assert params == {b'filename': 'foo-Ã¤.html'.encode()}


def test_attwithfnrawpctenca():
    dtype, params = parse_content_disposition(
        b'attachment; filename="foo-%41.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo-%41.html'}


def test_attwithfnusingpct():
    dtype, params = parse_content_disposition(
        b'attachment; filename="50%.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'50%.html'}


def test_attwithfnrawpctencaq():
    dtype, params = parse_content_disposition(
        rb'attachment; filename="foo-%\41.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo-%41.html'}


def test_attwithnamepct():
    dtype, params = parse_content_disposition(
        b'attachment; filename="foo-%41.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo-%41.html'}


def test_attwithfilenamepctandiso():
    dtype, params = parse_content_disposition(
        'attachment; filename="ä-%41.html"'.encode())
    assert dtype == b'attachment'
    assert params == {b'filename': 'ä-%41.html'.encode()}


def test_attwithfnrawpctenclong():
    dtype, params = parse_content_disposition(
        b'attachment; filename="foo-%c3%a4-%e2%82%ac.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo-%c3%a4-%e2%82%ac.html'}


def test_attwithasciifilenamews1():
    dtype, params = parse_content_disposition(
        b'attachment; filename ="foo.html"')
    assert dtype == b'attachment'
    assert params == {b'filename': b'foo.html'}



"""Tasty multipart form data parser built with cython."""
from pathlib import Path
from setuptools import setup, Extension

VERSION = (0, 1, 6)

setup(
    name="multifruits",
    version=".".join(map(str, VERSION)),
    description=__doc__,
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Pyrates",
    author_email="yohan.boniface@data.gouv.fr",
    url="https://github.com/pyrates/multifruits",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Environment :: Web Environment",
        "Development Status :: 4 - Beta",
    ],
    platforms=["POSIX"],
    license="MIT",
    ext_modules=[
        Extension(
            "multifruits",
            ["multifruits.c"],
            extra_compile_args=["-O3"],  # Max optimization when compiling.
        )
    ],
    provides=["multifruits"],
    include_package_data=True,
    extras_require={"dev": ["Cython==3.0.5"]}
)
